/*
 **********************************************************************
 * HISTORY
 * $Log:	allowed.c,v $
 * Revision 1.1  92/04/09  01:31:24  vac
 * Initial revision
 * 
 * Revision 1.6  89/06/18  15:10:43  gm0w
 * 	Extensively modified to be less wasteful using file I/O and memory.
 * 	[89/06/17            gm0w]
 * 
 * Revision 1.7  88/12/16  12:17:43  dimitris
 * 	Substituted realloc(dir,dirlen) to realloc(dir,dirlen+1) to
 * 	get the terminating null character.
 * 
 * Revision 1.5  88/12/13  14:42:55  dimitris
 * 	Modified allowed() to return 0 if the parent directory of the argument
 * 	is non local (i.e.  disallow rfs connections through ftp or paths like
 * 	/../dimitris/../dist.)
 * 
 * 17-Oct-88 Dimitris Varotsis (dimitris) at Carnegie Mellon University
 *	Created.
 *
 **********************************************************************
 */

#include <sys/param.h>
#include <stdio.h>
#include <libc.h>

extern int errno;

/*
 * allowed() will return 1 if arg is included in the file ".anonr" in its
 * parent or if any of the directories in its specification is in the
 * ".anonr" file of its parent directory.  (i.e. allowed (/usr/cs/dimitris)
 * will return 1 if "dimitris" is a line in /usr/cs/.anonr  or "cs" is a
 * line in /usr/.anonr or "usr" is a line in /.anonr .)
 */
allowed(arg)
char *arg;
{
    char opath[MAXPATHLEN], npath[MAXPATHLEN], lcomp[MAXPATHLEN];
    char *ltoken, *p;

    errno=0;

    /* Always allow the empty string */
    if (*arg == '\0')
	return(1);

    /* No remote command execution */
    if (*arg == '|')
	return(0);

    /* Always allow "." */
    if (*arg == '.' && arg[1] == '\0')
	return(1);

    if (getwd(opath) == NULL)
	return(0);

    /* Always make path an absolute filename */
    if (*arg != '/') {
	if (concat(npath, sizeof(npath), opath, "/", arg, NULL) == NULL)
	    return(0);
    } else
	if (concat(npath, sizeof(npath), arg, NULL) == NULL)
	    return(0);

    /*
     * ltoken:
     *   the last filename in the path (i.e. /usr/tmp/file ltoken="file")
     */
    ltoken = rindex(npath, '/');
    ltoken++;
    if (*ltoken != '\0' &&
	strcmp(ltoken, "..") != 0 && strcmp(ltoken, ".") != 0) {
	*(ltoken-1) = '\0';
	(void) strcpy(lcomp, ltoken);
	ltoken = lcomp;
    } else
	ltoken = NULL;

    /*
     * Use chdir() and getwd() to find where dir points to
     * (that will take care of ".." and symbolic links that use rfs),
     * and chdir() back to where you were (path.)
     */
    if (chdir(npath) != 0) {
	(void) chdir(opath);
	errno = 0;
	return(0);
    }
    if (getwd(npath) == NULL) {
	(void) chdir(opath);
	errno = 0;
	return(0);
    }
    (void) chdir(opath);

    /* paths in rfs not allowed */
    if (npath[0] == '/' && npath[1] == '.' && npath[2] == '.')
        return(0);

    p = npath + strlen(npath);
    if (ltoken == NULL) {
	/* avoid looking in /../.anonr for "/" */
	if (p == npath + 1)
	    return(0);
	do {
	    *(p+7) = *p;
	} while (*--p != '/');
	p++;
    } else {
	if (p != npath + 1)
	    *p++ = '/';
	(void) strcpy(p+7, ltoken);
    }

    /*
     * Loop trying each parent checking for its immediate child.
     * The buffer is split-shifted so that the path to the the
     * .anonr file is immediately followed by the component to
     * search for.
     */
    for (;;) {
	(void) bcopy(".anonr", p, 7);
	if (filefound(npath, p+7)) {
	    errno = 0;
	    return(1);
	}
	if (p == npath + 1)
	    break;
	*(--p) = '\0';
	while (*p != '/') {
	    *(p+7) = *p;
	    p--;
	}
	p++;
    }
    errno = 0;
    return(0);
}

/*
 * The next procedure opens the file anonr and then looks at each line
 * of this file to see if it is identical to the argument fnam in which
 * case it returns 1
 */
static
filefound(anonr, fnam)
char *anonr, *fnam;
{
    char buf[MAXPATHLEN];
    char *p;
    FILE *fp;

    if ((fp = fopen(anonr, "r")) == NULL)
	return(0);
    while (fgets(buf, sizeof(buf), fp) != NULL) {
	if (index(buf, '/') != NULL)
	    continue;
	if ((p = index(buf, '\n')) != NULL)
	    *p = '\0';
        if (strcmp(buf, fnam) != 0)
	    continue;
	fclose(fp);
	return(1);
    }
    fclose(fp);
    return(0);
}
