/*
 *  User Datagram Protocol Package
 **********************************************************************
 * HISTORY
 * 16-Apr-87  Glenn Marcy (gm0w) at Carnegie-Mellon University
 *	Added missing & before fromLen in recvfrom call.
 *
 **********************************************************************
 */

#include <errno.h>
#include <udp/udp.h>

#include <libc.h> /* for malloc() */

extern errno;

int udpdebug = 0;
#define	dprintf(flags)	if (udpdebug&(flags)) printf
#define	dif(flags)	if (udpdebug&(flags))
#define dreturn(val)	dprintf( UDP_TRACE)( "%s==>%d\n", rname, val); return val



/*
 *****************************************************************
 * Print failure diagnostic.
 *****************************************************************
 */
static void
udpfail( rname)
char *rname;
  {
  fprintf( stderr, "libudp: ");
  perror( rname);
  fflush( stderr);
  }



/*
 ****************************************************************
 * Return new udp object, or NULL on failure, with errno set.
 ****************************************************************
 */
udpObject
udpopen( lAddr, lPort, rAddr, rPort, recvTO, sendTO)
struct in_addr 	lAddr;
u_short 	lPort;
struct in_addr	rAddr;
u_short 	rPort;
long		recvTO;
long		sendTO;
  {
  static char * rname = "udpopen";
  int i;
  int bindval;
  int gsnval;
  int sourcelen;
  udpObject uO;
  int newsocket;

  dprintf( UDP_TRACE)( "%s( la:%d, lp:%d, ra:%d, rp:%d, rto:%d, sto:%d);\n",
		       rname,
		       lAddr.s_addr,
		       lPort,
		       rAddr.s_addr,
		       rPort,
		       recvTO,
		       sendTO);

  uO = (udpObject)malloc( sizeof( struct udpObject));

  if (uO == NULL)
    {
    errno = ENOBUFS;
    udpfail( rname);
    dreturn( NULL);
    }

  newsocket = socket( AF_INET, SOCK_DGRAM, 0);
  if (newsocket == -1)
    {
    udpfail( rname);
    free( uO);
    dreturn( NULL);
    }

  /* Just fill in enough to see if the bind will work. */
  bzero( &(uO->source), sizeof( struct sockaddr_in));
  uO->source.sin_family = AF_INET;
  uO->source.sin_port = lPort;
  uO->source.sin_addr = lAddr;

  bindval = bind( newsocket,
		&(uO->source),
		  sizeof(struct sockaddr_in));
  if (bindval == -1)
    {
    udpfail( rname);
    free( uO);
    dreturn( NULL);
    }

  uO->socket = newsocket;
  uO->sendTimeout.tv_sec = sendTO;
  uO->sendTimeout.tv_usec = 0;
  uO->recvTimeout.tv_sec = recvTO;
  uO->recvTimeout.tv_usec = 0;
  uO->destination.sin_family = AF_INET;
  uO->destination.sin_port = rPort;
  uO->destination.sin_addr = rAddr;

  /* Try to fill in with address assigned, if it was wildcarded */
  sourcelen = sizeof( uO->source);
  gsnval = getsockname( uO->socket, &(uO->source), &sourcelen);
  if (gsnval == -1)
    {
    udpfail( rname);
    free( uO);
    dreturn( NULL);
    }

  dreturn( uO);
  }



/*
 *****************************************************************
 * Close, free, etc. an udpObject.  No return value.
 *****************************************************************
 */
void
udpclose( uO)
udpObject uO;
  {
  static char *rname = "udpclose";

  dprintf( UDP_TRACE)( "%s();\n", rname);
  if (uO != NULL)
    {
    close( uO->socket);
    free( uO);
    }
  else
    {
    errno = EFAULT;
    udpfail( rname);
    }
  dprintf( UDP_TRACE)( "%s==>void.\n", rname);
  }



/*
 *****************************************************************
 * Returns the number of bytes written, -1 if an error occurred
 * (errno is set), or 0 if the write timed out.
 *****************************************************************
 */
int
udpwrite( uO, msg, len)
udpObject uO;
char *msg;
int len;
  {
  static char *rname = "udpwrite";

  dprintf( UDP_TRACE)( "%s( o:%d, msg:%d, len:%d);\n",
  		       rname, uO, msg, len);
  if (uO == NULL)
    {
    errno = EFAULT;
    udpfail( rname);
    dreturn( -1);
    }

  if ((uO->sendTimeout.tv_sec >= 0) &&
      (uO->sendTimeout.tv_usec >= 0))
    {
    int mask = 1 << (uO->socket);
    int selval = select( uO->socket + 1, 0, &mask, 0, &uO->sendTimeout);
    if (selval == -1)
      {
      dreturn( -1);  /* return errno set by select */
      }
    if (selval == 0)
      {
      dreturn( 0); /* Legitimate timeout */
      }
    if (selval != 1)
      {
      fprintf( stderr,
               "%s: select found more channels than asked for: %d\n",
	       rname,
	       selval);
      errno = EWOULDBLOCK;
      dreturn( -1);
      }
    }

  /*
   * Ok, if we get to here, either blocking writes are desired, or
   * select told us we wouldn't block.
   */
  {
  int returnval = sendto( uO->socket,
  			  msg,
			  len,
			  0,
			  &(uO->destination),
			  sizeof( struct sockaddr_in));
  dreturn( returnval);
  }
  }



/*
 *****************************************************************
 * Returns number of bytes read, -1 if an error occurred (errno
 * is set), or 0 if the read timed out.  If fromAddr is zero, no
 * from address will be returned.
 *****************************************************************
 */
int
udpread( uO, msg, len, fromAddr)
udpObject uO;
char *msg;
int len;
struct sockaddr_in *fromAddr;
  {
  static char * rname = "udpread";

  struct sockaddr_in temp;
  int fromLen = sizeof( struct sockaddr_in);

  dprintf( UDP_TRACE)( "%s( o:%d, msg: %d, len:%d, fA:%d)\n",
  		       rname,
		       uO,
		       msg,
		       len,
		       fromAddr);

  if (uO == NULL)
    {
    errno = EFAULT;
    udpfail( rname);
    dreturn( -1);
    }

  if (fromAddr == NULL)
    fromAddr = &temp;

  if ((uO->recvTimeout.tv_sec >= 0) &&
      (uO->recvTimeout.tv_usec >= 0))
    {
    int mask = 1 << (uO->socket);
    int selval = select( uO->socket+1, &mask, 0, 0, &uO->recvTimeout);
    if (selval == -1)
      {
      dreturn( selval);
      }
    if (selval == 0)
      {
      dreturn( 0); /* Legitimate timeout */
      }
    if (selval != 1)
      {
      fprintf( stderr, "%s: select found more channels than asked for: %d\n",
      	       rname,
	       selval);
      errno = EWOULDBLOCK;
      dreturn( -1);
      }
    }

  /*
   * If we get to here, either blocking receives are desired, or select
   * told us we wouldn't block.
   */
  {
  int returnval = recvfrom( uO->socket,
			    msg,
			    len,
			    0,
			    fromAddr,
			    &fromLen);
  dreturn( returnval);
  }
  }



/*
 *****************************************************************
 * Internet socket dump to a FILE *.
 *****************************************************************
 */
void
printSockaddr_in( skin, s)
struct sockaddr_in skin;
FILE * s;
  {
  u_char * ptr = (u_char *)&skin.sin_addr;
  fprintf( s,
	  "family %d, address %03u.%03u.%03u.%03u, port %d",
	  skin.sin_family,
	  ptr[ 0], ptr[ 1], ptr[ 2], ptr[ 3],
	  ntohs(skin.sin_port));
  }



/*
 *****************************************************************
 * Print a somewhat readable interpretation of an udpObject
 * onto a FILE *.
 *****************************************************************
 */
void
udpobjectdump( uO, s)
udpObject uO;
FILE * s;
  {
  if (uO == NULL)
    fprintf( s, "<null object>\n");
  else
    {
    fprintf( s, "socket: %d\n", uO->socket);
    fprintf( s, "   receive    timeout: %d\n", uO->recvTimeout.tv_sec);
    fprintf( s, "   send       timeout: %d\n", uO->sendTimeout.tv_sec);
    fprintf( s, "   source      socket: ");
    printSockaddr_in( uO->source, s);
    putc( '\n', s);
    fprintf( s, "   destination socket: ");
    printSockaddr_in( uO->destination, s);
    putc( '\n', s);
    }
  }
