
/* Copyright (c) 1992 Vincent Cate
 * All Rights Reserved.
 *
 * Permission to use and modify this software and its documentation
 * is hereby granted, provided that both the copyright notice and this
 * permission notice appear in all copies of the software, derivative works
 * or modified versions, and any portions thereof, and that both notices
 * appear in supporting documentation.  This software or any derivate works
 * may not be sold or distributed without prior written approval from
 * Vincent Cate.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND VINCENT CATE DISCLAIMS ALL
 * WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS.  IN NO EVENT SHALL
 * VINCENT CATE BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT, OR
 * CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS
 * OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE
 * OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE
 * OR PERFORMANCE OF THIS SOFTWARE.
 *
 * Users of this software agree to return to Vincent Cate any improvements
 * or extensions that they make and grant Vincent Cate the rights to
 * redistribute these changes.
 *
 *
 * Jan 1993 this file completely rewritten by
 *
 *           Juergen Hannken-Illjes (hannken@eis.cs.tu-bs.de)
 *
 * Exports:
 *
 * extern int
 * InTransit(char *pathName);
 *  return:
 *    TRUE:  if transfer of file is in progress
 *    FALSE: otherwise
 *
 *
 * extern int NumFinishedFTPs
 *  Number of succesful FTPs done so far
 *
 *
 * extern int
 * DoAnFtp(struct HostEntry *hostPtr, char *command, char *remPath,
 *         char *resultPath);
 *  return:
 *    AOK:       if file or directory has been transferred successfully
 *    AOLD:      if file doesn't exist on remote host (any more)
 *                 this doesn't work for directorys, as the error is written
 *                 to the output file as '... not found'
 *    AFAILED:   if transfer failed and will likely fail, if command is redone
 *    AWORKING:  if transfer is in progress or no free slot for another FTP
 *    SOFTERROR: if there has been some sort of FTP error that might go away with retry
 *
 *  side effect:
 *    uses and updates the following fields of *hostPtr:
 *      Path:       used to create hostname
 *      Type:       used to determine host type, set, if type changes
 *      HowWorking: used and set to collect recent failures
 *      LastUsed:   used and set
 *
 *    updates NumFinishedFTPs
 *
 *
 * Imports:
 *
 * SetLogBaseName()
 * debugLog
 * Log()
 * ATypeToString()
 *
 * PathToHostName()
 * TimeInSeconds()
 *
 * LeaveAnError()
 * RemoveError()
 *
 * Whenever you access a file in alex-cache, do it as follows:
 *
 *   if(ExistsInCacheAndSizeOk(pathname)) ||
 *           (InTransit(pathname) && BigEnough(pathname, NeededSize)) {
 *             use the file
 *    } else {
 *
 *         result = DoAnFtp(...);
 *         if(result == AOK)
 *           update stat info, proceed;
 *         if(result == AFAILED || result == AOLD)
 *           update stat info, return(AFAIL);
 *         if(result == SOFTERROR)
 *           return(AWORKING); (* try again next time *)
 *         if(result == AWORKING){
 *           if(SizeOk(pathname))
 *             proceed;
 *           else
 *             return(AWORKING);
 *       }
 *   }
 *   proceed as normal
 *
 */

#include "alexftp.h"

char passWord[MAXPATH];                 /* anon password: "user@host" */

/* parent code: the following code will be used from ALEX */

int NumFinishedFTPs = 0;		/* number of finished FTP's */

static struct ftptab{			/* structure for a child */
  int  inUse;				/* if 0, this child doesn't exist */
  int  lastUsed;			/* time of last status update */
  struct HostEntry *hostPtr;		/* host description for this child */
  char hostName[MAXPATH];		/* name of remote host */
  int  pid;				/* pid of child */
  int  cntrlIn;				/* pipe to child ("r") */
  int  cntrlOut;			/* pipe to child ("w") */
  char curCommand[MAXCOMMAND];		/* current command or ' ??', if idle */
  char resultPath[MAXPATH];		/* name of local file or '\0??', -"- */
  char lastError[MAXPATH];		/* last error, if any */
  char status;				/* current status of child (FTP_???) */
  double startTime;                     /* for performance statistics        */
}ftpCache[MAXFTPSOPEN];



RemoveError(ResultPath)
char *ResultPath;
{
    char AlexErrorPath[MAXPATH], DirPath[MAXPATH];

    if (strlen(ResultPath) > CACHEDIRLEN) {
        PathToDir(ResultPath, DirPath);

        (void) strcpy(AlexErrorPath, DirPath);
        (void) strcat(AlexErrorPath, SLASHALEXERROR);  /* add /.alex.error                  */
        (void) unlink(AlexErrorPath);                  /* if there we remove it not so what */
    }
}


/* Certain types of chars are really painful in filenames so remove them
 */
OnlyPathSafeChars(s)
char *s;
{
    if ((s==NULL) || (*s==0)) return;

    while (*s != 0) {
        if (isalnum(*s) || HasChar(".-=+_", *s)) {
            s++;                                   /* ok move to next */
        } else {
            *s++ = ' ';                            /* not ok make space then move to next */
        }
    }
}



/*  Make a .alex.error for FtpNum
 *       Find directory from .ResultPath
 *       Use .LastBack for text of error message
 */
int LeaveAnError(ResultPath, LastBack)
char *ResultPath, *LastBack;
{
    char DirPath[MAXPATH], ErrorFilePath[MAXPATH], ErrorMessage[MAXPATH];
    char AlexInfoPath[MAXPATH];
    int Result;
    FILE *AIfile;
    static int LastTimeAdded= -2;
    static char LastPathAdded[MAXPATH];
    int    MTime;

    if (strlen(ResultPath) <= CACHEDIRLEN) {
        ToLog(DBERROR, "LeaveAnError bogus ResultPath %s\n", ResultPath);
        return(AFAIL);
    }

    ToLog(6, "LeaveAnError %s %s\n", ResultPath, LastBack);

    PathToDir(ResultPath, DirPath);

    (void) strcpy(ErrorFilePath, DirPath);
    (void) strcat(ErrorFilePath, SLASHALEXERROR);

    (void) strcpy(ErrorMessage, "~ ALEX FTP ERROR  ");        /* so sorts to end of list */
    (void) strcat(ErrorMessage, LastBack);

    OnlyPathSafeChars(ErrorMessage);            /* skip the open quote */
    NoTrailingSpaces(ErrorMessage);             /* skip the open quote */

    Result=StringIntoFile(ErrorFilePath, ErrorMessage);

    if (Result == AOK) {
        (void) strcpy(AlexInfoPath, DirPath);
        (void) strcat(AlexInfoPath, SLASHALEXINFO);

        MTime = MTimeOfFile(AlexInfoPath);
        if (MTime < 0) {
            Log2("LeaveAnError not adding to .alex.info since non existant", AlexInfoPath);

        } else if (streql(LastPathAdded, AlexInfoPath) && (MTime == LastTimeAdded)) {
            Log2("LeaveAnError already added this error message to .alex.info ", ErrorMessage);

        } else {
            AIfile=AFOPEN(AlexInfoPath, "a");
            if (AIfile == NULL) {
                Log2("LeaveAnError could not open ", AlexInfoPath);
                Result=AFAIL;
            } else {
                Result=AddAnyErrorToStream(AIfile, DirPath);
                (void) AFCLOSE(AIfile);
            }
            strcpy(LastPathAdded, AlexInfoPath);
            LastTimeAdded=MTimeOfFile(AlexInfoPath);
        }
    }

    LogT("LeaveAnError returning ", Result);
    return(Result);
}






/*
 * create passWord
*
*
*static void
*CreatePassword()
*{
*  struct hostent *hp;
*  struct passwd *np;
*  char myname[MAXPATH];
*
*  np = getpwuid(getuid());
*
*  if(gethostname(myname,sizeof(myname)) < 0 ||
*	(hp = gethostbyname(myname)) == NULL)
*    sprintf(passWord,"%s@unkown_host",np ? np->pw_name : "unknown_user");
*  else
*    sprintf(passWord,"%s@%s",np ? np->pw_name : "unknown_user",hp->h_name);
*}
*/

/*
 * create name of a child, for logging purpose only
 */

static char *
NameOf(f)
     struct ftptab *f;
{
  static char buf[MAXPATH+10];

  sprintf(buf,"[%d:%s(%c)]",f-ftpCache,f->hostName,f->status);
  return(buf);
}

/*
 * kill this child so its slot may be reused
 */

static void
KillFtp(fp)
     struct ftptab *fp;
{
  struct ftptab *f;
  int pid;

  ToLog(1, "DoAnFtp::KillFtp: %s\n",NameOf(fp));

  kill(fp->pid,SIGHUP);
  while((pid = wait3(NULL,WNOHANG|WUNTRACED,NULL)) > 0)
    if(pid == fp->pid)
      break;
    else
      for(f = ftpCache; f < ftpCache+MAXFTPSOPEN; f++)
	if(f->inUse && f->pid == pid && FTP_RUNNING(f->status)){
	  ToLog(1, "DoAnFtp::KillFtp: %s did exit, status updated\n",NameOf(f));
	  f->status = FTP_FAILED;
	}

  if(fp->cntrlIn >= 0)
    close(fp->cntrlIn);
  if(fp->cntrlOut >= 0)
    close(fp->cntrlOut);
  bzero((char *)fp, sizeof(struct ftptab));
}

/*
 * create a new child
 */

static int
NewFtp(f,hostPtr,hostName,curCommand,resultPath, UidStr)
     struct ftptab *f;
     struct HostEntry *hostPtr;
     char *hostName;
     char *curCommand;
     char *resultPath;
     char *UidStr;
{
  int i,nfds;
  int pipeTo[2];
  int pipeFrom[2];

  if((hostPtr->HowWorking <= -2) &&                   /* 2 fast retries then slow */
	(hostPtr->LastUsed+(-(hostPtr->HowWorking)*SECSTILLRETRY) < TimeInSeconds())) {
      ToLog(DBFTP, "NewFtp  have failed with this host recently so just returning AFAIL %d\n ", 
                             hostPtr->HowWorking);
      return(AFAIL);                                  
  }

  f->inUse = 1;
  f->lastUsed = TimeInSeconds();
  f->hostPtr = hostPtr;
  f->hostPtr->LastUsed = TimeInSeconds();
  (void) strcpy(f->hostName,hostName);
  (void) strcpy(f->curCommand,curCommand);
  (void) strcpy(f->resultPath,resultPath);
  f->status = FTP_WORKING;
  f->cntrlIn = f->cntrlOut = -1;
  ToLog(DBFTP, "DoAnFtp::NewFtp: creating %s\n",NameOf(f));

  if(pipe(pipeFrom) < 0){
    ToLog(1, "DoAnFtp::NewFtp: cannot setup pipes\n");
    KillFtp(f);
    return(AFAIL);
  }
  if(pipe(pipeTo) < 0){
    close(pipeFrom[0]);
    close(pipeFrom[1]);
    ToLog(1, "DoAnFtp::NewFtp: cannot setup pipes\n");
    KillFtp(f);
    return(AFAIL);
  }
 
/* 
 * if(passWord[0] == '\0')
 *   CreatePassword();
 */
  (void) strcpy(passWord, UidStr);

  f->pid = fork();
    
  if(f->pid == 0){				/* child */
    debugLog = NULL;
    if(dup2(pipeTo[0],0) < 0 || dup2(pipeFrom[1],1) < 0) {
      ToLog(1, "DoAnFtp::NewFtp: dup2 failed\n");
      exit(1);
    }
    /* setlinebuf(stdout);                         not portable to HPUX so we fflush()  */
    nfds = getdtablesize();
    for(i = 2; i < nfds; i++)
      close(i);
    StartFtpClient(f->hostName);
    exit(1);
  }else if(f->pid > 0){			/* parent */
    close(pipeFrom[1]);
    close(pipeTo[0]);
    f->cntrlIn = pipeFrom[0];
    f->cntrlOut = pipeTo[1];
    f->lastError[0]=0;
    write(f->cntrlOut,curCommand,strlen(curCommand));
    f->startTime = TimeInSeconds();
    f->status = FTP_WORKING;
    ToLog(DBRPC, "DoAnFtp::NewFtp: created %s pid is %d\n",NameOf(f),f->pid);
    return(AWORKING);
  }else{
    KillFtp(f);
    return(AFAIL);
  }
  return(AFAIL);   /* for lint only */
}

/*
 * update status for this child
 */

static void
UpdateStatus(f)
     struct ftptab *f;
{
  char buf[MAXPATH+4];
  char lastStatus;
  int nread,i,done;

  ToLog(DBFTP, "DoAnFtp::UpdateStatus: %s", NameOf(f));

  done = 0;
  while(!done){
    lastStatus = f->status;
    if(ioctl(f->cntrlIn, FIONREAD, &nread) < 0 || nread < 2){
      done = 1;
      continue;
    }

    for(i = 0; i < nread; i++){
      read(f->cntrlIn,buf+i,1);
      if(buf[i] == '\n')
	break;
    }
    buf[i] = '\0';                                            

    ToLog(DBFTP, " {%s} ", buf);                   /* send everything we read to log */

    f->lastUsed = TimeInSeconds();

    ToLog(DBFTP, "+'%c'", buf[0]);
  
    if((f->status = buf[0]) == FTP_CONNECTED){
      ToLog(DBFTP, "(%s)", buf+1);
      if(buf[1] == SYS_UNIX) {
        f->hostPtr->Type = UNIX;
      } else if(buf[1] == SYS_VMS) {
        f->hostPtr->Type = DECVMS;                 
      } else {
        f->hostPtr->Type = UNKNOWN;                 
      }
    }

    switch(f->status){
      case FTP_CONNECTED:
        f->hostPtr->LastUsed = TimeInSeconds();
        f->status = FTP_WORKING;
        break;
      case FTP_OK:
        f->hostPtr->HowWorking = 0;
        f->hostPtr->LastUsed = TimeInSeconds();
        RemoveError(f->resultPath);
        f->lastError[0]=0;
        UpdatePerfStats(f);
        f->status = FTP_IDLE;
        NumFinishedFTPs++;
        break;
      case FTP_NOSUCHHOST:
        f->hostPtr->Type = NOSUCHHOST;
        (void) strcpy(f->lastError, "No such host");
        break;
      case FTP_NOANON:
        f->hostPtr->Type = NOANONFTP;
        (void) strcpy(f->lastError, "No anonymous FTP");
        break;
      case FTP_FAILED:
      case FTP_RETRY:
        if (strlen(buf)>2) {
          (void) strcpy(f->lastError, &buf[1]);               /* save error message from child */
        }
        f->hostPtr->HowWorking--;
        ToLog(DBFTP, "DoAnFtp::UpdateStatus HowWorking= %d\n", f->hostPtr->HowWorking);
        break;
      case FTP_NOTTHERE:
        (void) strcpy(f->lastError, "No such file");
        break;
      case FTP_EXITED:
        if(lastStatus == FTP_WORKING){
	  f->status = FTP_RETRY;
	  f->hostPtr->HowWorking--;
        }
        break;
    }
    ToLog(DBFTP, "='%c'", f->status);
  }

  ToLog(DBFTP, " done\n"); 
}

/*
 * build a command to be sent to child
 */

static int
BuildCommand(hostPtr,command,remPath,resultPath,curCommand)
     struct HostEntry *hostPtr;
     char *command;
     char *remPath;
     char *resultPath;
     char *curCommand;
{
  int len;

  while(remPath[0] == '/')                  /* remove any leading /  */
    remPath++;

  if(!strcmp(command,"get"))
    sprintf(curCommand,"%c ", FTP_GET);
  else if(!strcmp(command,"dir"))
    sprintf(curCommand,"%c ", FTP_DIR);
  else
    return(AFAIL);

  (void) strcat(curCommand,resultPath);
  (void) strcat(curCommand," ");

  if(remPath[0] == '\0' && curCommand[0] == FTP_GET){
    ToLog(DBFTP, "DoAnFtp::BuildCommand: cannot get empty path\n");
    return(AFAIL);
  }
  (void) strcat(curCommand,remPath);
  if(curCommand[0] == FTP_DIR){
    len = strlen(curCommand);
    if(curCommand[len-1] !='/' && remPath[0] != '\0') {   /* if name add a slash */
      curCommand[len++] = '/';
    }
    curCommand[len++] = '.';                              /* always add a dot    */
    curCommand[len] = '\0';

  }

  (void) strcat(curCommand,"\n");
  return(AOK);
}

/*
 * kill old childs
 */

static void
UpdateAllFtps()
{
  struct ftptab *f;
  char tmp[MAXFTPSOPEN+1];
  int i;

  for(i = 0; i < MAXFTPSOPEN; i++)
    tmp[i] = ftpCache[i].inUse ? 'X' : '-';
  tmp[MAXFTPSOPEN] = '\0';

  ToLog(DBFTP, "DoAnFtp::UpdateAllFtps: starting [%s]\n",tmp);

  /* kill all ftp's that have exited or are not running since 2*CLIENTTIMEOUT */

  for(f = ftpCache; f < ftpCache+MAXFTPSOPEN; f++)
    if(f->inUse){
      UpdateStatus(f);
      if(f->status == FTP_EXITED || (!FTP_RUNNING(f->status) &&
	    (TimeInSeconds()-f->lastUsed) > 2*CLIENTTIMEOUT))
        KillFtp(f);
    }
}

double TotalBytesFTPed=0;
double TotalFTPTime=0;

UpdatePerfStats(f)
struct ftptab *f;
{
    double FtpTime, BytesPerSec, BytesFtped;

    BytesFtped = SizeOfCachedFile(f->resultPath);
    if (BytesFtped > 0) {
        TotalBytesFTPed += BytesFtped;
        FtpTime = TimeInSeconds() - f->startTime;
        TotalFTPTime += FtpTime;
        BytesPerSec = BytesFtped/FtpTime;
        ToLog(DBPERF, "DoAnFtp BytesFtped= %0.0f Time= %2.2f B/S= %0.0f TotalBytesFTPed= %0.0f \n",
        BytesFtped,  FtpTime, BytesPerSec, TotalBytesFTPed);
    } else {
        ToLog(DBPERF, "DoAnFtp missing result file. (%c) %s \n", f->status, f->resultPath);
    }
}

/*
 * main function, see description on top of file
 */

int
DoAnFtp(hostPtr, command, remPath, resultPath, UidStr)
     struct HostEntry *hostPtr;
     char *command;
     char *remPath;
     char *resultPath;
     char *UidStr;                 
{
  char hostName[MAXPATH];
  char curCommand[MAXCOMMAND];
  struct ftptab *f;
  int result, OneFtpAlready;

  if (hostPtr->Type == NOSUCHHOST || hostPtr->Type == NOANONFTP) {
    result = hostPtr->Type;
    if (hostPtr->Type == NOANONFTP) {
       LeaveAnError(resultPath, "No anonymous FTP");
    }
    goto done;
  }

  PathToHostName(hostPtr->Path, hostName);

  ToLog(DBRPC, "DoAnFtp: %s %s %s %s\n",hostName,command,remPath,resultPath);

  if((result = BuildCommand(hostPtr,command,remPath,resultPath,curCommand)) != AOK)
    goto done;

  /* Now it's time to update all states */

  UpdateAllFtps();

  /* First lets see, if we are already running this command */
  OneFtpAlready=0;
  for(f = ftpCache; f < ftpCache+MAXFTPSOPEN; f++) {
    if(streql(f->hostName, hostName)) {
      if (!streql(f->curCommand,curCommand)) {
        OneFtpAlready=1;
        if(f->status != FTP_IDLE) {
            ToLog(DBFTP, "DoAnFtp: already running another command to this site %s\n", hostName);
        }
      } else {
        ToLog(DBFTP, "DoAnFtp: this command already running\n");
        switch(f->status) {
        case FTP_OK:
        case FTP_IDLE:
          f->curCommand[0] = ' ';
          f->resultPath[0] = '\0';
          f->status=FTP_IDLE;
	  result = AOK;
	  goto done;
        case FTP_NOTTHERE:
  	  f->curCommand[0] = ' ';
	  f->resultPath[0] = '\0';
	  f->status = FTP_IDLE;
	  result = AOLD;
	  goto done;
        case FTP_FAILED:
	  LeaveAnError(f->resultPath, f->lastError);
	  KillFtp(f);
	  result = AFAIL;
	  goto done;
        case FTP_RETRY:
          if((hostPtr->HowWorking <= -2) &&    /* 2 fast retries then slow */
	      (hostPtr->LastUsed+(hostPtr->HowWorking*SECSTILLRETRY) < TimeInSeconds())) {
	    LeaveAnError(f->resultPath, f->lastError);
	    KillFtp(f);
	    result = AFAIL;                            /* do not retry now */
	    goto done;
	  }
          ToLog(DBFTP, "DoAnFtp: FTP_RETRY with HowWorking= %d\n", hostPtr->HowWorking);
	  KillFtp(f);                                  
	  break;
        case FTP_WORKING:
	  result = AWORKING;
	  goto done;
        case FTP_NOSUCHHOST:
          hostPtr->Type=NOSUCHHOST;
          result = NOSUCHHOST;
          goto done;
        case FTP_NOANON:
	  LeaveAnError(f->resultPath, "No Anonymous FTP");
          hostPtr->Type=NOANONFTP;
          result = NOANONFTP;
          goto done;
        default:
	  ToLog(1, "DoAnFtp: ERROR unexpected state '%c'\n",f->status);
          AlexAssert(FALSE);
        }
      }
    }
  }

  /* Lets see, if we have an ftp to this host, thats idle */

  for(f = ftpCache; f < ftpCache+MAXFTPSOPEN; f++) {
    if(!strcmp(f->hostName,hostName) && f->status == FTP_IDLE){
      ToLog(DBFTP, "DoAnFtp: reusing an existing connection\n");
      (void) strcpy(f->curCommand,curCommand);
      (void) strcpy(f->resultPath,resultPath);
      write(f->cntrlOut,curCommand,strlen(curCommand));
      f->startTime = TimeInSeconds();
      f->status = FTP_WORKING;
      result = AWORKING;
      goto done;
    }
  }

  /* Ok, we must start a new one */
  if (!OneFtpAlready) {
    for(f = ftpCache; f < ftpCache+MAXFTPSOPEN; f++) {
      if(!f->inUse){
        result = NewFtp(f,hostPtr,hostName,curCommand,resultPath, UidStr);
        goto done;
      }
    }
  }

  /* No entry available, quit all childs, that are idle */

  for(f = ftpCache; f < ftpCache+MAXFTPSOPEN; f++) {
    if(f->status == FTP_IDLE){
      char buf[3];

      sprintf(buf,"%c\n",FTP_QUIT);
      write(f->cntrlOut, buf, strlen(buf));
      f->status = FTP_QUIT;
    }
  }

  /* All slots are busy, no chance at this moment, try again later */

  result = AWORKING;                                     /* was SOFTERROR */

done:;
  ToLog(DBRPC, "DoAnFtp: result is %s\n", ATypeToString(result));
  return(result);
}

/*
 * main function, see description on top of file
 */

int
InTransit(resultPath)
     char *resultPath;
{
  struct ftptab *f;

  for(f = ftpCache; f < ftpCache+MAXFTPSOPEN; f++)
    if(!strcmp(f->resultPath,resultPath)){
      ToLog(DBFTP, "DoAnFtp::InTransit: %s IS in transit %s\n",resultPath,NameOf(f));
      return(1);
    }

  ToLog(DBFTP, "DoAnFtp::InTransit: %s IS NOT in transit\n",resultPath);
  return(0);
}

/* dummys called from ALEX but not required */
CloseMostAllFDs () {}
ClearFtpCacheEntry () {}
void KillOldFtps() {}


